import util
import time
import random

def runGame(initialState, agents, display = True, displayStats = True):
    """ 
        Run a game from initialState with the given set of agents.  
            initialState is a GameState
            agents is a Dictionary of agents, which maps player names to agent functions
                (see game-agents.py)
            display specifies whether the progress of the game should be printed or not.
            displayStats specifies whether statistics about agent functions should be printed or not.

        Returns a Dictionary of terminal utilities.
    """
    currentState = initialState
    utilities = currentState.utility()    
    
    while not utilities:
        if display: print currentState
        player = currentState.currentPlayer()
        agent = agents[player]
        print ""

        currentState.resetSearchStats()
        startTime = time.clock()
        
        agentAction = agent.getAction(currentState)
        
        if display: 
            print "Player %s agent returned action %s" % (player, agentAction)
        if displayStats: 
            (nact, nsucc) = currentState.getSearchStats()
            print "actions called %d times and successor called %d times in %s seconds" %  (nact, nsucc, time.clock()-startTime)
        
        assert agentAction in currentState.actions()
        currentState = currentState.successor(agentAction)
        utilities = currentState.utility()     
    
    print currentState
    print "\nGame finished with utilities", utilities
    return utilities






class GameState:
    """
    Abstract GameState class. Your classes should inherit from this class and override
    the methods below.  Calling the default constructor should return the initial state.
    """
    
    def getPlayers(self):
        """  Return a tuple of players for this game, with the player-to-move first """
        abstract
        
        
    def getPreviousMove(self):
        """  Return the previous move, or None if this is the initial game state """
        abstract
        
    
    def actions(self):
        """  Return a list of actions possible from this state. """
        abstract
        
        
    def successor(self, action):
        """  Return the successor state of self on action.  
             Precondition: action in actions(self) """
        abstract
    
    
    def utility(self):
        """  Returns None if the game is nonterminal, or a Dictionary 
             associating players to utilities if it is terminal. """
        abstract


    def __str__(self):
        """  Return a printable string representation of this state. """
        abstract
        

    def representation(self):
        """  Return a hashable representation of self that will be 
          used for representation and equality tests. 
          
          By default, uses the printable representation (may be very inefficient). """
        return self.__str__()


    # These methods are used to evaluate GameAgents    
    def getSearchStats(self):
        """ return (# actions calls, # successor calls) """
        abstract
    
    def resetSearchStats(self):
        """ Reset internal state of problem for tracking search statistics. """    
        abstract



     # You may find these shortcut methods useful:    
    def successors(self):
        """  Return a list of (action, successor) pairs """
        return [(action, self.successor(action)) for action in self.actions()]

    def shuffledSuccessors(self):
        """  Return a list of (action, successor) pairs """
        succ = self.successors()
        random.shuffle(succ)
        return succ
    

    def currentPlayer(self):
        """  Returns the player whose turn it is to move. """
        return self.getPlayers()[0]


    def currentPlayerUtility(self):
        """  Returns None if the game is nonterminal, or else a utility for the current player. """
        utility = self.utility()
        if utility:
            return utility[self.currentPlayer()]

    
    # Methods to ensure that equality and hashing work as expected.
    def __eq__(self, other):
        return self.representation() == other.representation()

    def __ne__(self, other):
        return self.representation() != other.representation()

    def __hash__(self):
        return hash(self.representation())





class FixedDepthHeuristicGameStateAdapter(GameState):
    """
        A wrapper class, which takes an ordinary GameState representing a 
        state in a two-player, zero-sum game of arbitrary depth, and returns a new 
        game state where actions and successors are identical, utilities for states 
        with remainingDepth > 0 are left unchanged, and all states at remainingDepth = 0 
        are considered terminal, where heuristicEvalFn(state) supplies the utility for
        currentPlayer().  
        
        This class allows you to use your exhaustive Minimax and AlphaBeta 
        search algorithms with a heuristic evaluation function without modifying them
        at all, by simply wrapping the game state in question in an instance of this class.
    """
    
    def __init__(self, state, heuristicEvalFn, remainingDepth):
        self.state = state
        self.heuristicEvalFn = heuristicEvalFn
        self.remainingDepth = remainingDepth

    def getPlayers(self):        return self.state.getPlayers()
    def getPreviousMove(self):   return self.state.getPreviousMove()
    def actions(self):           return self.state.actions()
    def __str__(self):           return str(self.state)
    def representation(self):    return (self.state.representation(), self.remainingDepth)
    def getSearchStats(self):    return self.state.getSearchStats()
    def resetSearchStats(self):  return self.state.resetSearchStates()

    def successor(self, action): 
        nextState = self.state.successor(action)
        nextDepth = self.remainingDepth-1
        return FixedDepthHeuristicGameStateAdapter(nextState, self.heuristicEvalFn, nextDepth)
    
    def utility(self):
        """  Returns None if the game is nonterminal, or a Dictionary 
             associating players to utilities if it is terminal. """
        actualUtility = self.state.utility()
        if actualUtility: return actualUtility

        if self.remainingDepth > 0: return None
                 
        (currentPlayer, otherPlayer) = self.getPlayers()
        playerScore = self.heuristicEvalFn(self.state)
        return {currentPlayer : playerScore, otherPlayer: -playerScore}



